import java.util.*;

/** 
 * Skeleton code for doubly-connected edge list data structure
 */ 
 
class DCEL {
 
  ArrayList<Vertex> vertices = new ArrayList<Vertex>();
  ArrayList<HalfEdge> halfEdges = new ArrayList<HalfEdge>();
  ArrayList<Face> faces = new ArrayList<Face>();
  
  Face outerFace = null;
  
  public DCEL() {
    outerFace = this.new Face();
  }
  
  /**
   * Creates a DCEL from a polygon given as a counter-clockwise list of 
   * points. 
   */
  public DCEL(ArrayList<Point3d> polygon) {
    outerFace = this.new Face();
    Face innerFace = this.new Face();
    for (Point3d p : polygon) { newVertex(p); }
    for (Vertex v : vertices) { newHalfEdge(v, innerFace); }
    int n = halfEdges.size();
    for (int i = 0; i < n; i++) {
      HalfEdge curr = halfEdges.get(i),
               next = halfEdges.get((i+1) % n);
      curr.makeNext(next);
      HalfEdge twin = newHalfEdge(next.origin, outerFace);
      curr.makeTwins(twin);
    }
    for (int i = 0; i < n; i++) {
      HalfEdge curr = halfEdges.get(i);
      curr.twin.makeNext(curr.prev.twin);
    }
  }
  
  Vertex newVertex(Point3d coordinates) { return this.new Vertex(coordinates); }
  HalfEdge newHalfEdge(Vertex origin, Face face) { return this.new HalfEdge(origin, face); }
  Face newFace() { return this.new Face(); }
  
  ArrayList<HalfEdge> getCycleFrom(HalfEdge e) {
    HalfEdge loopEdge = e;
    ArrayList<HalfEdge> retEdges = new ArrayList<HalfEdge>();
    do {
      retEdges.add(loopEdge);
      loopEdge = loopEdge.next;
    } while (loopEdge != e);
    return retEdges; 
  }
  
  ArrayList<HalfEdge> getCycleAround(HalfEdge e) {
    HalfEdge loopEdge = e;
    ArrayList<HalfEdge> retEdges = new ArrayList<HalfEdge>();
    do {
      retEdges.add(loopEdge);
      loopEdge = loopEdge.twin.next;
    } while (loopEdge != e);
    return retEdges;
  }
  
  /**
   * Splits the face incident to e1 and e2 by adding an edge between the origins of e1 and e2. 
   * If e1 and e2 are not incident to the same face, or e1 is the next or prev of e2, nothing should
   * happen (because the call isn't "legal"). 
   */ 
  void splitFace(HalfEdge e1, HalfEdge e2) {
    // 0. Check whether the call is legal and return if not. 
    if (e1.face != e2.face || e1.next == e2 || e1.prev == e2 || e1 == e2) {
      return;
    }
    // 1. Create a new face for the second half of the split. 
    Face newFace = this.newFace();
    
    // 2. Create new half-edges for the splitting edge:
    HalfEdge s1 = this.new HalfEdge(e2.origin, e1.face);
    HalfEdge s2 = this.new HalfEdge(e1.origin, newFace);
    s1.makeTwins(s2);
    
    // 3. Insert your new half-edges into the cyclic lists: 
    e2.prev.makeNext(s1);
    e1.prev.makeNext(s2);
    s1.makeNext(e1);
    s2.makeNext(e2);
    
    // 4. Make sure that all the half-edges point to the right face
    for (HalfEdge e : getCycleFrom(e2)) {
      e.face = newFace;
    }
    
    // 5. Make sure that each face correctly refers to an incident half-edge
    newFace.incidentEdge = e2;
    e1.face.incidentEdge = e1;
  }
  
  class HalfEdge {
    
    HalfEdge twin = null, next = null, prev = null; 
    Vertex origin = null; 
    Face face = null;
    
    private HalfEdge(Vertex origin, Face face) {
      this.origin = origin;
      this.face = face;
      halfEdges.add(this);
      origin.outgoingEdge = this;
      face.incidentEdge = this;
    }
    
    public double length() {
      return origin.coord.minus(next.origin.coord).norm();
    }
    
    Vertex getDestination() {
      return twin.origin;
    }
    
    /**
     * Helper for making this the twin of that and that the twin of this. 
     */
    void makeTwins(HalfEdge that) {
      this.twin = that; 
      that.twin = this;
    }
    
    /** 
     * Helper for making that the next of this and this the prev of that. 
     */
    void makeNext(HalfEdge that) {
      this.next = that;
      that.prev = this;
    }
     
    Vector3d bisector() {
      return new Vector3d(1,0,0);
    }
    
    Point3d pointAlongBisectorAtHeight(double t) {
      // Change this line: 
      Point3d insetPoint = this.origin.coord;
      return insetPoint.plus(new Vector3d(0,0,t));
    }
    
    double contractionHeight() {
      return 100.0;
    }
    
    double angle() {
      Vector3d eprev = prev.origin.coord.minus(origin.coord);
      Vector3d enext = next.origin.coord.minus(origin.coord);
      return eprev.thetaTo(enext);
    }
    
    void contract() {
      ArrayList<HalfEdge> ccw = getCycleAround(this);
      this.prev.makeNext(this.next);
      this.twin.prev.makeNext(this.twin.next);
      halfEdges.remove(this);
      halfEdges.remove(this.prev);
      this.face.incidentEdge = this.prev;
      this.twin.face.incidentEdge = this.twin.prev;
      for (HalfEdge he : ccw) {
        if (he != this) he.makeOrigin(this.getDestination());
      }
      vertices.remove(this.origin);
    }
    
    void makeOrigin(Vertex o) {
      this.origin = o;
      o.outgoingEdge = this;
    }
  }
  
  class Vertex {
    Point3d coord;
    HalfEdge outgoingEdge; 
    private  Vertex(Point3d coordinates) {
      vertices.add(this);
      this.coord = coordinates; 
    }
  }
  
  class Face {
    HalfEdge incidentEdge; 
    private Face() {
      faces.add(this);
    }
    
    void offset(double t) {
      ArrayList<HalfEdge> edges = DCEL.this.getCycleFrom(incidentEdge);
      
      // Create the bisector edges
      for (HalfEdge he : edges) {
        Vertex newV = newVertex(he.pointAlongBisectorAtHeight(t));
        HalfEdge out = newHalfEdge(he.origin, this);
        HalfEdge in = newHalfEdge(newV, this);
        out.makeTwins(in);
        he.prev.makeNext(out);
        in.makeNext(he);
        out.makeNext(in);
      }
      
      // Cap off the faces
      for (HalfEdge he : edges) {
        Face newFace = newFace();
        HalfEdge new1 = newHalfEdge(he.next.getDestination(), newFace);
        HalfEdge new2 = newHalfEdge(he.prev.origin, this);
        
        new1.makeTwins(new2);
        
        he.prev.prev.makeNext(new2);
        new2.makeNext(he.next.next);
        
        new1.makeNext(he.prev);
        he.next.makeNext(new1);
        
        he.prev.face = newFace;
        he.face = newFace;
        he.next.face = newFace;
      }
    }
    
    double calcContractionHeight() {
      return 100;
    }
    
    HalfEdge shortestHalfEdge() {
      HalfEdge shortest = this.incidentEdge;
      // find shortest here
      return shortest;
    }
    
    boolean isDegenerate() {
      return incidentEdge.next == incidentEdge.prev;
    }
    
    HalfEdge removeDegenerate() {
      if (!this.isDegenerate()) return null;
      incidentEdge.twin.makeTwins(incidentEdge.next.twin);
      halfEdges.remove(incidentEdge);
      halfEdges.remove(incidentEdge.next);
      faces.remove(this);
      return incidentEdge.twin;
    }
  }
  
}